# This file is part of SOUPDE.
#
# SOUPDE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SOUPDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SOUPDE.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2010 Matthieu Weber <matthieu.weber@jyu.fi>

import random, numpy

inf = float("inf")

class Individual(object):
  _no = 0

  def __init__(self, data):
    self.data = data
    self.fitness = inf
    self.no = Individual._no
    Individual._no += 1

  def __str__(self):
    return str(self.no)

  def id(self):
    return "%08x" % (0x100000000 + id(self))

  def get_fitness(self):
    return self.fitness

class ArrayIndividual(Individual):

  def __init__(self, data, sigma = 0.1, F = 0.7, Cr = 0.3):
    Individual.__init__(self, data)
    self.sigma = sigma
    self.F = F
    self.Cr = Cr
    self.d = 0
    self.Gcc = 0
    self.mask = None
  
  @classmethod
  def create_random(cls, size, interval = (-1,1), sigma = 0.1):
    return cls(numpy.array([random.uniform(*interval) for i in range(size)]), sigma=sigma)

class Population(object):

  def __init__(self, pop_size, ind_type, ind_size, interval = (-1,1)):
    self.pop_size = pop_size
    self.ind_type = ind_type
    self.ind_size = ind_size
    self.interval = interval
 
  def __str__(self):
    return " ".join([str(i) for i in self.individuals])

  def set_individuals(self, individuals):
    self.individuals = individuals

  def get_best(self):
    if len(self.individuals):
      return min(map(lambda x: (x.get_fitness(),x), self.individuals[0:self.pop_size]))[1]
    else:
      return None

  def get_best_index(self):
    return min(map(lambda x: (x[1].get_fitness(),x[0]), enumerate(self.individuals[0:self.pop_size])))[1]

class ArrayPopulation(Population):
  def __init__(self, pop_size, ind_size, interval = (-1,1), sigma = 0.1):
    Population.__init__(self, pop_size, ArrayIndividual, ind_size, interval)
    self.sigma = sigma
    self.individuals = [self.ind_type.create_random(ind_size, self.interval, self.sigma) for i in range(self.pop_size)]

  def create_random(self):
    return [self.ind_type.create_random(self.ind_size, self.interval, self.sigma) for i in range(self.pop_size)]

  def __str__(self):
    return "[" + ", ".join(["% 4d: %6.3f" % (_.no, _.fitness) for _ in self.individuals]) + "]"

class Incubator(object):
  def __init__(self, pop, calc_fitness):
    self.fitness_evaluations = 0
    self.calc_fitness = calc_fitness
    self.pop = pop

  def get_pop_size(self):
    return self.pop.pop_size

class MultiIncubator(Incubator):
  """A base class for multi-population algorithm.
  Not meant for direct implementation but for deriving new classes from it.
  The create() class method is a factory function for the algorithm.
  """
  def __init__(self, pop, calc_fitness):
    Incubator.__init__(self, pop, calc_fitness)

  @classmethod
  def generate_population(self, seed, size, function):
    """Factory function for generating an initial population.
    Parameters:
      seed: seed for the random number generator in the population
            initialization step
      size: size of the population
      function: fitness function
    """
    state = random.getstate()
    random.seed(seed)
    pop = [ArrayPopulation(size[0], function.dimension, function.boundaries) for i in range(size[1])]
    random.setstate(state)
    return pop
 
  @classmethod
  def create(cls, seed, size, function, **kwargs):
    """Factory function for an algorithm
    Parameters:
      seed: seed for the random number generator in the population
            initialization step
      size: size of the population
      function: fitness function
      kwargs: any additional keyword argument
    """
    pop = cls.generate_population(seed, size, function)
    return cls(pop, function, **kwargs)

  def get_pop_size(self):
    return self.incubators[0].pop.pop_size * len(self.incubators)

  def get_best(self):
    return min([(i.pop.get_best().get_fitness(),i) for i in  self.incubators[:]])[1]

class DEexp(Incubator):
  """DE/rand/1/exp"""
  name = "DE/rand/1/exp"

  def __init__(self, pop, calc_fitness, Cr = 0.3, F = 0.7, init = True):
    """Initialization of a DE/rand/1/exp algorithm
    Parameters:
      pop: population as generated by the generate_population() class method
      calc_fitness: fitness function
      Cr: crossover rate
      F: scale factor
      init: flag, the population's fitness values are calculated if True
    """
    Incubator.__init__(self, pop, calc_fitness)
    self.best_changed = False
    self.Cr = Cr
    self.F = F
    if init:
      for i in self.pop.individuals:
        i.fitness = self.calc_fitness(i.data)
        self.fitness_evaluations += 1

  @classmethod
  def generate_population(self, seed, size, function):
    """Factory function for generating an initial population.
    Parameters:
      seed: seed for the random number generator in the population
            initialization step
      size: size of the population
      function: fitness function
    """
    state = random.getstate()
    random.seed(seed)
    pop = ArrayPopulation(size, function.dimension, function.boundaries)
    random.setstate(state)
    return pop

  @classmethod
  def create(cls, seed, size, function, **kwargs):
    """Factory function for an algorithm
    Parameters:
      seed: seed for the random number generator in the population
            initialization step
      size: size of the population
      function: fitness function
      kwargs: any additional keyword argument
    pop = cls.generate_population(seed, size, function)
    return cls(pop, function, **kwargs)
  """  
  def params(self):
    return {'s_pop': self.pop.pop_size, 'Cr': self.Cr, 'F': self.F}

  def get_others(self, current):
    sample = random.sample(range(self.pop.pop_size), 3)
    while sample[0] == sample[1] or sample[0] == sample[2] or sample[1] == sample[2] \
       or sample[0] == current   or sample[1] == current   or sample[2] == current:
      sample = random.sample(range(self.pop.pop_size), 3)
    return [self.pop.individuals[i] for i in sample]

  def iterate(self):
    children = []
    for i in xrange(self.pop.pop_size):
      x = self.get_others(i)
      jrand = random.randrange(self.pop.ind_size)
      choices = numpy.array([False] * self.pop.ind_size)
      choices[jrand] = True
      for _ in range(1, self.pop.ind_size):
        if random.uniform(0,1) <= self.Cr:
          choices[(_ + jrand) % self.pop.ind_size] = True
        else:
          break
      diff = x[1].data - x[2].data
      cand = self.calc_fitness.domain_rollover(x[0].data + self.F * diff)
      u = self.pop.ind_type(choices.choose(self.pop.individuals[i].data, cand))
      u.fitness = self.calc_fitness(u.data)
      self.fitness_evaluations += 1
      success = u.fitness < self.pop.individuals[i].fitness
      if success:
        children.append(u)
      else:
        children.append(self.pop.individuals[i])
    self.pop.set_individuals(children + self.pop.individuals[self.pop.pop_size:])
    best_fitness = self.pop.get_best().get_fitness()
    return self.fitness_evaluations, best_fitness

class SOUPDE(MultiIncubator):
  """Shuffle Or Refresh Parallel Differential Evolution"""
  name = "Shuffle Or Refresh Parallel Differential Evolution"

  def __init__(self, pop, calc_fitness, Cr = 0.9, p_s = 0.01, p_u = 0.01):
    """Initializes a SOUPDE object
    Parameters:
      pop: a population as generated by the class' generate population
      calc_fitness: a fitness function
      Cr: Crossover Rate for the Differential Evolution algorithm
      p_s: shuffling probability
      p_u: update probability
    """
    MultiIncubator.__init__(self, pop, calc_fitness)
    self.Cr = Cr
    self.p_s = p_s
    self.p_u = p_u
    self.incubators = [DEexp(p, calc_fitness, Cr, random.uniform(0.1,1)) for p in pop]

  def params(self):
    return {'s_pop': [self.incubators[0].pop.pop_size, len(self.incubators)], 'Cr': self.Cr, 'p_s': self.p_s, 'p_u': self.p_u}

  def shuffle(self):
    all = []
    for p in [inc.pop for inc in self.incubators]:
      for i in p.individuals:
        all.append(i)
    random.shuffle(all)
    i = 0
    for p in [inc.pop for inc in self.incubators]:
      for ind in range(len(p.individuals)):
        p.individuals[ind] = all[i]
        i += 1

  def iterate(self):
    """Runs one generation of the algorithm"""
    if random.uniform(0,1) < self.p_s:
      self.shuffle()
    if random.uniform(0,1) < self.p_u:
      for inc in self.incubators:
        inc.F = random.uniform(0.1,1)
    results = [incubator.iterate() for incubator in self.incubators]
    self.fitness_evaluations = sum([x[0] for x in results])
    best_fitness = min([x[1] for x in results])
    return self.fitness_evaluations, best_fitness

