# This file is part of SOUPDE.

# SOUPDE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SOUPDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SOUPDE.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2010 Matthieu Weber <matthieu.weber@jyu.fi>

# Functions F1 to F6 are reimplemented from the benchmark defined by
# K. Tang, X. Yao, P. N. Suganthan, C. MacNish, Y. P. Chen, C. M. Chen, and Z.
# Yang. Benchmark Functions for the CEC'2008 Special Session and Competition
# on Large Scale Global Optimization. Technical Report, Nature Inspired
# Computation and Applications Laboratory, USTC, China, 2007.

# Functions F7 to F19star are reimplemented from the benchmark defined by
# F. Herrera, M. Lozano and D. Molina in Test Suite for the Special Issue of
# Soft Computing on Scalability of Evolutionary Algorithms and other
# Metaheuristics for Large Scale Continuous Optimization Problems
# February 9, 2010

import numpy, f_data

dtype=numpy.float_
if hasattr(numpy,'float128'):
  dtype = numpy.float128
elif hasattr(numpy,'float96'):
  dtype = numpy.float96

class TestFunction:
  def __init__(self, dimension, boundaries = None):
    self.dimension = dimension
    #print "dim:",self.dimension
    #print "bou:",self.boundaries
    if boundaries:
      self.boundaries = boundaries
    self.low_boundaries = numpy.array([self.boundaries[0]] * dimension)
    self.high_boundaries = numpy.array([self.boundaries[1]] * dimension)
    self.diagonal = numpy.sqrt(numpy.add.reduce((self.high_boundaries-self.low_boundaries)**2))
 
  def id(self):
    return self.__class__.__name__

  def domain_clamp(self, x):
    return numpy.maximum(numpy.minimum(x,self.high_boundaries),self.low_boundaries)

  def domain_rollover(self, x):
    return ((x - self.low_boundaries) % (self.high_boundaries - self.low_boundaries)) + self.low_boundaries

  #def domain_mirror(self, x):
  #  x - 

  def domain_infinite(self, x):
    if numpy.logical_and.reduce((x < self.low_boundaries) | (x > self.high_boundaries)):
      return float("inf")
    return x

  def __call__(self, x):
    if numpy.logical_and.reduce((x < self.low_boundaries) | (x > self.high_boundaries)):
      return float("inf")
    return self.call(x)

class F1(TestFunction):
  """Shifted Sphere"""
  name = "Shifted Sphere"
  boundaries = (-100, 100)
  def call(self, x):
    #print x, ":", f_data.f1[:self.dimension]
    return numpy.add.reduce((x-f_data.f1[:self.dimension])**2)
    
class F2(TestFunction):
  """Schwefel Problem"""
  name = "Schwefel Problem"
  boundaries = (-100, 100)
  def call(self, x):
    return numpy.max(numpy.abs(x-f_data.f2[:self.dimension]))

class F3(TestFunction):
  """Shifted Rosenbrock"""
  name = "Shifted Rosenbrock"
  boundaries = (-100, 100)
  def call(self, x):
    z = x - f_data.f3[:self.dimension] + 1
    return numpy.add.reduce((100 * (numpy.roll(z,-1) - z**2) ** 2 + (1-z)**2)[:self.dimension-1])

class F4(TestFunction):
  """Shifted Rastrigin"""
  name = "Shifted Rastrigin"
  boundaries = (-5, 5)
  def call(self, x):
    z = x - f_data.f4[:self.dimension]
    return 10 * z.size + numpy.add.reduce(z**2 - 10 * numpy.cos(2*numpy.pi*z))

class F5(TestFunction):
  """Shifted Griewangk"""
  name = "Shifted Griewangk"
  boundaries = (-600, 600)
  def call(self, x):
    z = x - f_data.f5[:self.dimension]
    i = numpy.arange(1,z.size+1)
    return numpy.add.reduce(z**2)/4000 - numpy.prod(numpy.cos(z / numpy.sqrt(i))) + 1
  
class F6(TestFunction):
  """Shifted Ackley"""
  name = "Shifted Ackley"
  boundaries = (-32, 32)
  def call(self, x, a=20, b=0.2, c=2*numpy.pi):
    z = x - f_data.f6[:self.dimension]
    n = z.size 
    return -a * numpy.exp(-b * numpy.sqrt(numpy.add.reduce(z**2) / n)) - numpy.exp(numpy.add.reduce(numpy.cos(c * z)) / n) + a + numpy.e

class F7(TestFunction):
  """Schwefel's Problem 2.22"""
  name = "Schwefel's Problem 2.22"
  boundaries = (-10, 10)
  def call(self, x):
    z = numpy.abs(x)
    return numpy.add.reduce(z) + numpy.prod(z, dtype=dtype)

class F8(TestFunction):
  """Schwefel's Problem 1.2"""
  name = "Schwefel's Problem 1.2"
  boundaries = (-65.536, 65.536)
  def call(self, x):
    return numpy.add.reduce(numpy.add.accumulate(x)**2)

class F9(TestFunction):
  """Extended f_10"""
  name = "Extended f_10"
  boundaries = (-100,100)

  def call(self, x):
    x0 = x
    x1 = numpy.roll(x,-1)
    p = (x0**2 + x1**2)
    return numpy.add.reduce( p**0.25 * (numpy.sin(50*(p**0.1))**2 + 1))
    
class F10(TestFunction):
  """Bohachevsky"""
  name = "Bohachevsky"
  boundaries = (-15,15)

  def call(self,x):
    x0 = x[:-1]
    x1 = numpy.roll(x,-1)[:-1]
    return numpy.add.reduce(x0**2 + 2*x1**2 - 0.3 * numpy.cos(3*numpy.pi*x0) - 0.4 * numpy.cos(4*numpy.pi*x1) + 0.7)

class F11(TestFunction):
  """Schaffer"""
  name = "Schaffer"
  boundaries = (-100, 100)
  def call(self,x):
    x0 = x[:-1]
    x1 = numpy.roll(x,-1)[:-1]
    p = (x0**2 + x1**2)
    return numpy.add.reduce( p**0.25 * (numpy.sin(50*(p**0.1))**2 + 1))

def shift(function, shift_data):
  """Shift a given function"""
  class shiftedFunction(function):
    name = function.name + "_shifted"
    def __call__(self, x):
      z = x - shift_data[:self.dimension]
      return function.__call__(self, z)
  shiftedFunction.__name__ = function.__name__ + "_shifted"
  return shiftedFunction

F7_shifted = shift(F7, f_data.f7)
F8_shifted = shift(F8, f_data.f8)
F9_shifted = shift(F9, f_data.f9)
F10_shifted = shift(F10, f_data.f10)
F11_shifted = shift(F11, f_data.f11)

def FCombine(class1,class2,mns):
    """ Hybrid composition function from 
        Test Suite for the Special Issue of Soft
        Computing on Scalability of Evolutionary
        Algorithms and other Metaheuristics for Large
        Scale Continuous Optimization Problems """
    class classComb(TestFunction):
        name = class1.name+"-"+class2.name+"-"+str(mns)
        (x1,y1) = class1.boundaries
        (x2,y2) = class2.boundaries
        boundaries = (max(x1,x2),min(y1,y2))
        def __call__(self,x):
            #print "call bds",self.boundaries,len(x)
            even_elements,odd_elements = splitDomain(self.dimension,mns,x)
            length_p1=len(even_elements)#int(self.dimension*mns)
            length_p2=len(odd_elements)#self.dimension-length_p1
            #print "sizes:",length_p1,length_p2
            #print "lens:",len(even_elements),len(odd_elements)
            testfunction1 = class1(length_p1)
            testfunction2 = class2(length_p2)
            return testfunction1(even_elements) + testfunction2(odd_elements)
                
    classComb.__name__ = class1.__name__ + "-" + class2.__name__+"-"+str(mns)
    return classComb

def splitDomain(dim,mns,x):
            if mns < 0.5:
                length_p1 = int(dim*mns)
                even_elements = x[0::2][:length_p1] 
                odd_elements  = numpy.concatenate((x[1::2][:length_p1]
                                                  ,x[2*length_p1:]))
            else:
                length_p2 = int(dim*(1-mns))
                odd_elements = x[1::2][:length_p2] 
                even_elements = numpy.concatenate((x[0::2][:length_p2]
                                                  ,x[2*length_p2:]))
            return even_elements,odd_elements

F12=FCombine(F9,F1,0.25)
F12.__name__ = "F12"
F13=FCombine(F9,F3,0.25)
F13.__name__ = "F13"
F14=FCombine(F9,F4,0.25)
F14.__name__ = "F14"
F15=FCombine(F10,F7,0.25)
F15.__name__ = "F15"

F15_shifted = shift(F15, f_data.f15)

F16star=FCombine(F9,F1,0.5)
F16star.__name__ = "F16*"
F17star=FCombine(F9,F3,0.75)
F17star.__name__ = "F17*"
F18star=FCombine(F9,F4,0.75)
F18star.__name__ = "F18*"
F19star=FCombine(F10,F7,0.75)
F19star.__name__ = "F19*"

F19star_shifted = shift(F19star, f_data.f19)
